
#include <Pokitto.h>
#include <clock_11u6x.h>
#include <timer_11u6x.h>
#include "testing123.h"

using Pokitto::Core;
using Pokitto::Display;
using Pokitto::Buttons;

// Timer match number select
constexpr unsigned int MATCHNUM = 1;
// Sample rate
constexpr unsigned int FREQUENCY = 22050;

unsigned int ticks = 0;

inline void enableDAC() {
    volatile unsigned int* PIO1 = (unsigned int*)(0x40044060);
    volatile unsigned int* PIO2 = (unsigned int*)(0x400440f0);
    volatile unsigned int* DIR1 = (unsigned int*)(0xa0002004);
    volatile unsigned int* DIR2 = (unsigned int*)(0xa0002008);
    PIO1[28] = PIO1[29] = PIO1[30] = PIO1[31] = 1 << 7;
    PIO2[20] = PIO2[21] = PIO2[22] = PIO2[23] = 1 << 7;
    *DIR1 |= (1 << 28) | (1 << 29) | (1 << 30) | (1 << 31);
    *DIR2 |= (1 << 20) | (1 << 21) | (1 << 22) | (1 << 23);
    
    volatile unsigned int* SET1 = (unsigned int*)(0xa0002204);
    *DIR1 |= 1 << 17;
    *SET1 = 1 << 17;
}

inline void writeDAC(unsigned char out) {
    volatile unsigned char* P1 = (unsigned char*)(0xa0000020);
    volatile unsigned char* P2 = (unsigned char*)(0xa0000040);
    P1[28] = out & 1; out >>= 1;
    P1[29] = out & 1; out >>= 1;
    P1[30] = out & 1; out >>= 1;
    P1[31] = out & 1; out >>= 1;
    P2[20] = out & 1; out >>= 1;
    P2[21] = out & 1; out >>= 1;
    P2[22] = out & 1; out >>= 1;
    P2[23] = out;
}

extern "C" void TIMER32_0_IRQHandler() {
    volatile unsigned int* TIMER = (unsigned int*)(0x40014000);
    if((TIMER[0] & (1 << MATCHNUM)) != 0) {
        TIMER[0] = 1 << MATCHNUM;
        
        if(ticks < sizeof(testing123_snd)) {
            writeDAC(testing123_snd[ticks]);
        }
        ++ticks;
        if(ticks >= 5 * FREQUENCY) {
            // Reset after 5 seconds
            ticks = 0;
        }
    }
}

// Custom timer initialization code
inline void enableTimerCustom(unsigned int number, unsigned int matchnum, unsigned int frequency) {
    Display::print("Custom\n");
    
    volatile unsigned int* SYSCON = (unsigned int*)(0x40048000);
    SYSCON[32] |= (number == 0) ? (1 << 9) : (1 << 10); // Enable clock for the timer
    unsigned int msel = (SYSCON[2] & 0x1f) + 1;         // Get feedback divider value
    unsigned int timerFreq = 12000000 * msel / SYSCON[30];
    
    constexpr unsigned int TIMER_0 = 0x40014000, TIMER_1 = 0x40018000;
    volatile unsigned int* TIMER = (unsigned int*)((number == 0) ? TIMER_0 : TIMER_1);
    TIMER[1] = 1 << 1;                           // Reset timer
    TIMER[6 + matchnum] = timerFreq / frequency; // Set match value (timer's duration)
    TIMER[5] |= 1 << (3 * matchnum);             // Generate interrupt when timer equals match value
    TIMER[5] |= 1 << (3 * matchnum + 1);         // Reset timer on match
    TIMER[1] = 1 << 0;                           // Enable the timer counter
    
    // Enable timer interrupt
    constexpr unsigned char TIMER_0_IRQ = 18, TIMER_1_IRQ = 19;
    volatile unsigned int* NVICTL = (unsigned int*)(0xe000e100);
    NVICTL[0] = (number == 0) ? (1 << TIMER_0_IRQ) : (1 << TIMER_1_IRQ);
}

// Timer initialization code copied directly from PokittoLib
inline void enableTimerPokittoLib(unsigned int matchnum, unsigned int frequency) {
    Display::print("PokittoLib\n");
    
    Chip_TIMER_Init(LPC_TIMER32_0);
    /* Timer rate is system clock rate */
    unsigned int timerFreq = Chip_Clock_GetSystemClockRate();
    /* Timer setup for match and interrupt at TICKRATE_HZ */
    Chip_TIMER_Reset(LPC_TIMER32_0);
    /* Enable both timers to generate interrupts when time matches */
    Chip_TIMER_MatchEnableInt(LPC_TIMER32_0, matchnum);
    /* Setup 32-bit timer's duration (32-bit match time) */
    Chip_TIMER_SetMatch(LPC_TIMER32_0, matchnum, (timerFreq / frequency));
    /* Setup both timers to restart when match occurs */
    Chip_TIMER_ResetOnMatchEnable(LPC_TIMER32_0, matchnum);
    /* Start both timers */
    Chip_TIMER_Enable(LPC_TIMER32_0);
    
	/* Clear both timers of any pending interrupts */
	constexpr unsigned int TIMER_32_0_IRQn = 18;
	NVIC_ClearPendingIRQ((IRQn_Type)TIMER_32_0_IRQn);
    /* Redirect IRQ vector - Jonne*/
    NVIC_SetVector((IRQn_Type)TIMER_32_0_IRQn, (uint32_t)&TIMER32_0_IRQHandler);
	/* Enable both timer interrupts */
	NVIC_EnableIRQ((IRQn_Type)TIMER_32_0_IRQn);
}

// PokittoLib timer initialization without IRQ redirect
inline void enableTimerPokittoLibNoRedir(unsigned int matchnum, unsigned int frequency) {
    Display::print("PokittoLib, no IRQ redirect\n");
    
    Chip_TIMER_Init(LPC_TIMER32_0);
    /* Timer rate is system clock rate */
    unsigned int timerFreq = Chip_Clock_GetSystemClockRate();
    /* Timer setup for match and interrupt at TICKRATE_HZ */
    Chip_TIMER_Reset(LPC_TIMER32_0);
    /* Enable both timers to generate interrupts when time matches */
    Chip_TIMER_MatchEnableInt(LPC_TIMER32_0, matchnum);
    /* Setup 32-bit timer's duration (32-bit match time) */
    Chip_TIMER_SetMatch(LPC_TIMER32_0, matchnum, (timerFreq / frequency));
    /* Setup both timers to restart when match occurs */
    Chip_TIMER_ResetOnMatchEnable(LPC_TIMER32_0, matchnum);
    /* Start both timers */
    Chip_TIMER_Enable(LPC_TIMER32_0);
    
    // Enable timer interrupt
	constexpr unsigned int TIMER_32_0_IRQn = 18;
    volatile unsigned int* NVICTL = (unsigned int*)(0xe000e100);
    NVICTL[0] = 1 << TIMER_32_0_IRQn;
}

int main() {
    Core::begin();
    Display::persistence = true;
    
    enableDAC();
    
    Core::update();
    Display::print("Choose enableTimer method\nA - Custom\nB - PokittoLib\nC - PokittoLib, no redir\n\n");
    Core::update();
    
    // Wait until button A is no longer pressed
    do {
        Buttons::update();
    } while(Buttons::aBtn());
    
    // Loop until user chooses A, B or C
    while(true) {
        Buttons::update();
        if(Buttons::pressed(BTN_A)) {
            enableTimerCustom(0, MATCHNUM, FREQUENCY);
            break;
        }
        else if(Buttons::pressed(BTN_B)) {
            enableTimerPokittoLib(MATCHNUM, FREQUENCY);
            break;
        }
        else if(Buttons::pressed(BTN_C)) {
            enableTimerPokittoLibNoRedir(MATCHNUM, FREQUENCY);
            break;
        }
    }
    
    Display::print("Playing...\n");
    
    while(Core::isRunning()) {
        if(Core::update()) {
            
        }
    }
    
    return 0;
}
